
/*
 * LIB/NEWSFEED.C
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

Prototype int FeedAdd(const char *msgid, History *h, const char *nglist, const char *npath, int size);
Prototype void FeedFlush(void);
Prototype void LoadNewsFeed(History *h, int force);
Prototype void TouchNewsFeed(void);
Prototype int IsFiltered(const char *hlabel, const char *nglist);

Prototype FILE *FeedFo;

typedef struct GroupList {
    struct GroupList	*gl_Next;
    char		*gl_Label;
    struct Node		*gl_GroupAcceptBase;
} GroupList;

typedef struct NewsFeed {
    struct NewsFeed	*nf_Next;
    char		*nf_Label;
    struct Node		*nf_PathAliasBase;
    struct Node		*nf_GroupAcceptBase;
    struct Node		*nf_FilterBase;
    int			nf_MaxCrossPost;	/* 0=unlimited */
    int			nf_MaxArtSize;		/* 0=unlimited */
    int			nf_MaxPathLen;		/* maximum ouging path len */
} NewsFeed;

NewsFeed 	*NFBase;
GroupList	*GLBase;
NewsFeed 	*NFCache;

char	NFBuf[16384];
int	NFIdx;
FILE	*FeedFo;

int feedQueryPaths(NewsFeed *feed, const char *npath, int size);
int feedQueryGroups(NewsFeed *feed, const char *nglist);
int feedGroupQuery(Node *list, const char *group, int accept);
int filterQueryGroups(NewsFeed *feed, const char *nglist);
int filterGroupQuery(NewsFeed *feed, const char *group);
int feedPathQuery(NewsFeed *feed, const char *path);

void resolveGroupList(const char *label, Node *scan);
void AltFeed(NewsFeed *nf, FILE *fi);
int TooNear(time_t t);

/*
 * FEEDADD() 
 */

int 
FeedAdd(const char *msgid, History *h, const char *nglist, const char *npath, int size)
{
    NewsFeed *nf;
    int cnt = 0;
    int r = 0;

    LoadNewsFeed(h, 0);

    for (nf = NFBase; nf; nf = nf->nf_Next) {
	if (FeedFo && 
	    feedQueryPaths(nf, npath, size) == 0 && 
	    feedQueryGroups(nf, nglist) == 0
	) {
	    char path[256];

	    ++cnt;
	    ArticleFileName(path, h, 0);

	    fprintf(FeedFo, "FEED %s %s %ld,%ld %s\n",
		nf->nf_Label,
		path,
		h->boffset,
		h->bsize,
		msgid
	    );
	    if (DebugOpt > 1) {
		printf("FEED %s %s %s %ld,%ld\n",
		    nf->nf_Label,
		    path,
		    msgid,
		    h->boffset,
		    h->bsize
		);
	    }
	} else {
	    if (DebugOpt > 1) {
		printf("%s SKIP (%d,%d) %s\t(%s)\t(%s)\n", 
		    nf->nf_Label,
		    feedQueryPaths(nf, npath, size),
		    feedQueryGroups(nf, nglist),
		    msgid,
		    npath,
		    nglist
		);
	    }
	}
    }
    if (FeedFo) {
	fprintf(FeedFo, "STATS c=%d s=%d m=%s\n",
	    cnt,
	    size,
	    msgid
	);
	fflush(FeedFo);	/* simplify the parent reader on the pipe */
	if (ferror(FeedFo)) {
	    syslog(LOG_CRIT, "lost backchannel to master server");
	    r = -1;
	}
    }
    return(r);
}

void
FeedFlush(void)
{
/*    NewsFeed *nf;*/

    /*
    if (FeedFo)
	fflush(FeedFo);
    */
}

int 
IsFiltered(const char *hlabel, const char *nglist)
{
    NewsFeed *nf;
    int r = 0;

    if (hlabel[0] == 0)
	hlabel = "ME";

    if ((nf = NFCache) == NULL) {
	for (nf = NFBase; nf; nf = nf->nf_Next) {
	    if (strcmp(hlabel, nf->nf_Label) == 0)
		break;
	}
	if (nf == NULL)
	    nf = (NewsFeed *)-1;
	NFCache = nf;
    }

    if (nf == (NewsFeed *)-1) {
	r = -1;
    } else {
	r = filterQueryGroups(nf, nglist);
    }
    if (DebugOpt)
	printf("IsFiltered: %d (%s,%s)\n", r, hlabel, nglist);
    return(r);
}

/*
 * Return 0 if the feed does NOT have an alias for an element in the path
 *	AND the article is not too large, otherwise return -1.
 */

int
feedQueryPaths(NewsFeed *feed, const char *npath, int size)
{
    const char *p;
    int cnt = 0;

    if (feed->nf_MaxArtSize && size > feed->nf_MaxArtSize)
	return(-1);

    while (*npath == ' ' || *npath == '\t')
	++npath;
    for (p = npath; p; p = strchr(p, '!')) {
	char pat[MAXGNAME];
	int l;

	++cnt;

	if (*p == '!')
	    ++p;
	for (l = 0; l < MAXGNAME - 1 && p[l] && p[l] != '!' && p[l] != ' ' && p[l] != '\t' && p[l] != '\n' && p[l] != '\r'; ++l)
	    ;
	strncpy(pat, p, l);
	pat[l] = 0;
	if (feedPathQuery(feed, pat) == 0)
	    break;
    }
    if (p == NULL) {
	/*
	 * no path aliases matched, we are ok... unless MaxPathLen 
	 * is set.
	 */
	if (feed->nf_MaxPathLen && feed->nf_MaxPathLen < cnt)
	    return(-1);
	return(0);
    }
    /*
     * path alias found, skip this feed
     */

    return(-1);
}

/*
 * Return 0 if the feed has a group match against a group in the group list
 * Return -1 if there is no match
 */

int
feedQueryGroups(NewsFeed *feed, const char *nglist)
{
    const char *p;
    int r = -1;
    int count = 0;

    while (*nglist == ' ' || *nglist == '\t')
	++nglist;
    for (p = nglist; p && r != -2; p = strchr(p, ',')) {
	int l;
	int nr;
	char group[MAXGNAME];

	if (*p == ',')
	    ++p;

	for (l = 0; l < MAXGNAME - 1 && p[l] && p[l] != ',' && p[l] != ' ' && p[l] != '\t' && p[l] != '\n' && p[l] != '\r'; ++l)
	    ;

	/*
	 * r:	0	feed based on group
	 *	-1	do not feed based on group
	 *	-2	do not feed based on group if group appears at all
	 */

	strncpy(group, p, l);
	group[l] = 0;
	nr = feedGroupQuery(feed->nf_GroupAcceptBase, group, -1);
	++count;
	if (nr != -1)
	    r = nr;
    }
    if (r == 0 && feed->nf_MaxCrossPost && count > feed->nf_MaxCrossPost)
	r = -1;
    return(r);
}

/*
 * Return 0 if a group in the group list has been
 * filtered from the particular incoming feed.
 */

int
filterQueryGroups(NewsFeed *feed, const char *nglist)
{
    const char *p;

    while (*nglist == ' ' || *nglist == '\t')
	++nglist;
    for (p = nglist; p; p = strchr(p, ',')) {
	int l;
	char group[MAXGNAME];

	if (*p == ',')
	    ++p;
	for (l = 0; l < MAXGNAME - 1 && p[l] && p[l] != ',' && p[l] != ' ' && p[l] != '\t' && p[l] != '\n' && p[l] != '\r'; ++l)
	    ;
	strncpy(group, p, l);
	group[l] = 0;
	if (filterGroupQuery(feed, group) == 0)
	    break;
    }
    if (p == NULL)
	return(-1);
    return(0);
}

/*
 * feedGroupQuery()
 *
 *	-2	do not feed ANY group if this group appears in the newsgroups
 *		line.
 *
 *	-1	do not feed this group
 *
 *	0	feed this group
 *
 */

int
feedGroupQuery(Node *node, const char *group, int accept)
{
    static int RecurCnt = 0;
    static int RecurWarn = 0;

    for (; node; node = node->no_Next) {
	if (node->no_Value == 2) {
	    if (node->no_Data) {
		/*
		 * GroupList recursion
		 */
		++RecurCnt;
		if (RecurCnt < 10) {
		    accept = feedGroupQuery(((GroupList *)node->no_Data)->gl_GroupAcceptBase, group, accept);
		} else {
		    if (RecurWarn == 0) {
			syslog(LOG_EMERG, "DNewsfeeds file grouplist infinite recursion at %s!\n", ((GroupList *)node)->gl_Label);
			RecurWarn = 1;
		    }
		}
		--RecurCnt;
	    }
	} else if (WildCmp(node->no_Name, group) == 0) {
	    if (node->no_Value < 0) {
		accept = node->no_Value;	/* -1 or -2 */
	    } else if (node->no_Value > 0) {
		accept = 0;			/* +1	    */
	    }
	}
    }
    return(accept);
}

int
filterGroupQuery(NewsFeed *feed, const char *group)
{
    Node *node;
    int accept = -1;

    for (node = feed->nf_FilterBase; node; node = node->no_Next) {
	if (WildCmp(node->no_Name, group) == 0) {
	    if (node->no_Value < 0) {
		accept = -1;
	    } else if (node->no_Value > 0) {
		accept = 0;
	    }
	}
    }
    return(accept);
}

int
feedPathQuery(NewsFeed *feed, const char *path)
{
    Node *node;
    int found = -1;

    for (node = feed->nf_PathAliasBase; node; node = node->no_Next) {
	if (WildCmp(node->no_Name, path) == 0) {
	    found = 0;
	    break;
	}
    }
    return(found);
}

unsigned long NFGmtMin = (unsigned long)-1;
time_t NFMTime = 0;

void
TouchNewsFeed(void)
{
    FILE *fi = xfopen("r+", "%s/dnewsfeeds", NewsHome);
    if (fi) {
	int c;
	if ((c = fgetc(fi)) != EOF) {
	    fseek(fi, 0L, 0);
	    fputc(c, fi);
	    fflush(fi);
	}
	fclose(fi);
    }
}

void
LoadNewsFeed(History *h, int force)
{
    int timeChanged = 0;
    /*
     * check for dnewsfeeds file modified once a minute
     */

    if (force || h == NULL || h->gmt != NFGmtMin) {
	struct stat st = { 0 };
	FILE *fi;

	timeChanged = 1;

	if (h)
	    NFGmtMin = h->gmt;

	fi = xfopen("r", "%s/dnewsfeeds", NewsHome);

	if (fi == NULL)
	    syslog(LOG_EMERG, "%s/dnewsfeeds file not found", NewsHome);

	if (force || fi == NULL || NFMTime == 0 ||
	    (fstat(fileno(fi), &st) == 0 && st.st_mtime != NFMTime &&
	    !TooNear(st.st_mtime))
	) {
	    char buf[MAXGNAME+256];

	    NFMTime = st.st_mtime;	/* may be 0 if file failed to open */

	    /*
	     * flush existing feed information
	     */

	    FeedFlush();

	    /*
	     * free up feed structures
	     */

	    {
		NewsFeed *nf;

		while ((nf = NFBase) != NULL) {
		    NFBase = nf->nf_Next;
		    FreeNodeList(&nf->nf_PathAliasBase);
		    FreeNodeList(&nf->nf_GroupAcceptBase);
		    FreeNodeList(&nf->nf_FilterBase);
		    free(nf);
		}
		NFCache = NULL;
	    }
	    {
		GroupList *gl;

		while ((gl = GLBase) != NULL) {
		    GLBase = gl->gl_Next;
		    FreeNodeList(&gl->gl_GroupAcceptBase);
		    free(gl);
		}
	    }

	    /*
	     * load up new feed structures
	     */

	    {
		NewsFeed *nf = NULL;
		GroupList *gl = NULL;
		Node **paNode = NULL;
		Node **pgNode = NULL;
		int lineNo = 0;
		FILE *altFi = NULL;

		while (fi && fgets(buf, sizeof(buf), fi) != NULL) {
		    char *s1 = strtok(buf, " \t\n");
		    char *s2 = strtok(NULL, " \t\n");
		    int err = 1;

		    ++lineNo;

		    if (s1 == NULL || *s1 == '#')
			continue;

		    if (strcmp(s1, "label") == 0) {
			if (nf || gl) {
			    syslog(LOG_CRIT, 
				"Config line %d, no end before new label!\n",
				lineNo
			    );
			    if (paNode)
				*paNode = NULL;
			    if (pgNode)
				*pgNode = NULL;
			    if (nf) {
				nf->nf_Next = NFBase;
				NFBase = nf;
				if (altFi) {
				    AltFeed(nf, altFi);
				    fclose(altFi);
				    altFi = NULL;
				}
			    }
			    if (gl) {
				gl->gl_Next = GLBase;
				GLBase = gl;
			    }
			    nf = NULL;
			    gl = NULL;
			}
			if (s2) {
			    char path[256];

			    nf = malloc(sizeof(NewsFeed) + strlen(s2) + 1);
			    memset(nf, 0, sizeof(NewsFeed));
			    nf->nf_Label = (char *)(nf + 1);
			    strcpy(nf->nf_Label, s2);
			    paNode = &nf->nf_PathAliasBase;
			    pgNode = &nf->nf_GroupAcceptBase;

			    sprintf(path, "%s/feeds/%s", NewsHome, s2);
			    altFi = fcdopen(path, "r");
			    err = 0;
			}
		    } else if (strcmp(s1, "alias") == 0) {
			if (nf && s2) {
			    Node *node = MakeNode(s2, 0);
			    *paNode = node;
			    paNode = &node->no_Next;
			    err = 0;
			}
		    } else if (strcmp(s1, "filter") == 0) {
			if (nf && s2) {
			    Node *node = MakeNode(s2, 1);

			    node->no_Next = nf->nf_FilterBase;
			    nf->nf_FilterBase = node;
			    err = 0;
			}
		    } else if (strcmp(s1, "maxcross") == 0) {
			if (nf && s2) {
			    nf->nf_MaxCrossPost = strtol(s2, NULL, 0);
			    err = 0;
			}
		    } else if (strcmp(s1, "maxpath") == 0) {
			if (nf && s2) {
			    nf->nf_MaxPathLen = strtol(s2, NULL, 0);
			    err = 0;
			}
		    } else if (strcmp(s1, "maxsize") == 0) {
			if (nf && s2) {
			    char *p;

			    nf->nf_MaxArtSize = strtol(s2, &p, 0);

			    switch(*p) {
			    case 'g':
			    case 'G':
				nf->nf_MaxArtSize *= 1024;
				/* fall through */
			    case 'm':
			    case 'M':
				nf->nf_MaxArtSize *= 1024;
				/* fall through */
			    case 'k':
			    case 'K':
				nf->nf_MaxArtSize *= 1024;
				break;
			    }
			    err = 0;
			}
		    } else if (strcmp(s1, "groupdef") == 0) {
			if (nf || gl) {
			    syslog(LOG_CRIT,
			       "Config line %d, no end before new groupdef!\n", 
				lineNo
			    );
			    if (paNode)
				*paNode = NULL;
			    if (pgNode)
				*pgNode = NULL;
			    if (nf) {
				nf->nf_Next = NFBase;
				NFBase = nf;
				if (altFi) {
				    AltFeed(nf, altFi);
				    fclose(altFi);
				    altFi = NULL;
				}
			    }
			    if (gl) {
				gl->gl_Next = GLBase;
				GLBase = gl;
			    }
			    nf = NULL;
			    gl = NULL;
			}
			if (s2) {
			    gl = malloc(sizeof(GroupList) + strlen(s2) + 1);
			    memset(gl, 0, sizeof(GroupList));
			    gl->gl_Label = (char *)(gl + 1);
			    strcpy(gl->gl_Label, s2);
			    paNode = NULL;
			    pgNode = &gl->gl_GroupAcceptBase;
			    err = 0;
			}
		    } else if (strcmp(s1, "groupref") == 0) {
			if (pgNode && s2 && altFi == NULL) {
			    Node *node = MakeNode(s2, 2);
			    *pgNode = node;
			    pgNode = &node->no_Next;
			}
			if (pgNode && s2)
			    err = 0;
		    } else if (strcmp(s1, "delgroupany") == 0) {
			if (pgNode && s2 && altFi == NULL) {
			    Node *node = MakeNode(s2, -2);
			    *pgNode = node;
			    pgNode = &node->no_Next;
			}
			if (pgNode && s2)
			    err = 0;
		    } else if (strcmp(s1, "delgroup") == 0) {
			if (pgNode && s2 && altFi == NULL) {
			    Node *node = MakeNode(s2, -1);
			    *pgNode = node;
			    pgNode = &node->no_Next;
			}
			if (pgNode && s2)
			    err = 0;
		    } else if (strcmp(s1, "addgroup") == 0) {
			if (pgNode && s2 && altFi == NULL) {
			    Node *node = MakeNode(s2, 1);
			    *pgNode = node;
			    pgNode = &node->no_Next;
			}
			if (pgNode && s2)
			    err = 0;
		    } else if (strcmp(s1, "end") == 0) {
			if (nf) {
			    *paNode = NULL;
			    *pgNode = NULL;
			    nf->nf_Next = NFBase;
			    NFBase = nf;
			    if (altFi) {
				AltFeed(nf, altFi);
				fclose(altFi);
				altFi = NULL;
			    }
			    nf = NULL;
			    err = 0;
			}
			if (gl) {
			    *pgNode = NULL;
			    gl->gl_Next = GLBase;
			    GLBase = gl;
			    gl = NULL;
			    err = 0;
			}
		    } else {
			syslog(LOG_CRIT, 
			    "Config line %d, unknown command\n", 
			    lineNo
			);
			err = 0;
		    }
		    if (err) {
			syslog(LOG_CRIT, "Config line %d, command in unexpected position or command requires argument\n", lineNo);
		    }
		}
		if (nf) {
		    *paNode = NULL;
		    *pgNode = NULL;
		    nf->nf_Next = NFBase;
		    NFBase = nf;
		    nf = NULL;
		}
		if (gl) {
		    *pgNode = NULL;
		    gl->gl_Next = GLBase;
		    GLBase = gl;
		    gl = NULL;
		}
		if (altFi) {
		    fclose(altFi);
		}
	    }
	}
	if (fi)
	    fclose(fi);

	/*
	 * Resolve GroupList references after the fact.  This allows
	 * grouplist definitions to put after newsfeed definitions.
	 */
	{
	    NewsFeed *nf;
	    GroupList *gl;

	    for (nf = NFBase; nf; nf = nf->nf_Next)
		resolveGroupList(nf->nf_Label, nf->nf_GroupAcceptBase);
	    for (gl = GLBase; gl; gl = gl->gl_Next)
		resolveGroupList(gl->gl_Label, gl->gl_GroupAcceptBase);
	}
    }
}

void
resolveGroupList(const char *label, Node *scan)
{
    for (; scan; scan = scan->no_Next) {
	if (scan->no_Value == 2) {
	    GroupList *gl;

	    for (gl = GLBase; gl; gl = gl->gl_Next) {
		if (strcmp(scan->no_Name, gl->gl_Label) == 0) {
		    scan->no_Data = gl;
		    break;
		}
	    }
	    if (gl == NULL) {
		syslog(LOG_CRIT, "Error: grouplist %s does not exist (from %s)\n", scan->no_Name, label);
	    }
	}
    }
}

void
AltFeed(NewsFeed *nf, FILE *fi)
{
    char buf[MAXGNAME+256];
    Node **pgNode = &nf->nf_GroupAcceptBase;

    while (*pgNode)
	pgNode = &(*pgNode)->no_Next;

    while (fgets(buf, sizeof(buf), fi) != NULL) {
	char *s1 = strtok(buf, " \t\n");
	char *s2 = strtok(NULL, " \t\n");

	if (s1 == NULL)
	    continue;
	if (s2 == NULL)
	    continue;
	if (strcmp(s1, "addgroup") == 0) {
	    Node *node = MakeNode(s2, 1);
	    *pgNode = node;
	    pgNode = &node->no_Next;
	} else if (strcmp(s1, "delgroup") == 0) {
	    Node *node = MakeNode(s2, -1);
	    *pgNode = node;
	    pgNode = &node->no_Next;
	} else if (strcmp(s1, "delgroupany") == 0) {
	    Node *node = MakeNode(s2, -2);
	    *pgNode = node;
	    pgNode = &node->no_Next;
	}
    }
    *pgNode = NULL;
}

int
TooNear(time_t t)
{
    time_t now = time(NULL);
    long dt = (long)(now - t);

    if (dt > -10 && dt < 10)
	return(1);
    return(0);
}

