
/*
 * DIABLO.C	INTERNET NEWS TRANSIT AGENT, By Matthew Dillon
 *
 * Diablo implements the news transfer portion of the INN command
 * set.  Basically ihave, check, takethis, mode stream, head, and stat.
 * Not much more.  The purpose is to transfer news as quickly and
 * as efficiently as possible.
 *
 * Diablo is a forking server.  It supports both standard non-streaming
 * ihave feeds and streaming check/takethis feeds.  In fact, it supports
 * streaming on its entire command set if you wish to use it that way.
 * It forks on every connection and is able to do history file lookups
 * and article reception in parallel.  History file updates use minimal
 * locking and take advantage of the flexibility of INN's return codes.
 *
 * There is no active file or newsgroups file, and Control's are not 
 * processed in any form (not even cancels).  DIABLO is strictly a
 * backbone redistribution/transit agent.
 *
 *
 * (c)Copyright 1997, Matthew Dillon, All Rights Reserved.  Refer to
 *    the COPYRIGHT file in the base directory of this distribution 
 *    for specific rights granted.
 */

#include "defs.h"

#if NEED_TERMIOS
#include <sys/termios.h>
#endif

typedef struct Feed {
    struct Feed *fe_Next;
    char	*fe_Label;
    int		fe_Fd;
    char	*fe_Buf;
    int		fe_BufIdx;
    int		fe_BufMax;
    int		fe_Failed;
} Feed;

typedef struct Retain {
    struct Retain *re_Next;
    FILE	*re_Fi;
    FILE	*re_Fo;
    int		re_What;
} Retain;

#define RET_CLOSE	1
#define RET_PAUSE	2
#define RET_LOCK	3

void DiabloServer(void);
void DoAccept(int lfd);
void DoPipe(int fd);
char *Authenticate(int fd, struct sockaddr_in *asin, char *label);
void DoSession(int fd);
void LogSession(void);
void LogSession2(void);
void DoCommand(int ufd);
int LoadArticle(FILE *fi, char *buf, const char *msgid, int noWrite);
void FlushArtFd(void);

void writeFeedOut(const char *label, const char *file, const char *msgid, const char *offSize);
void flushFeeds(int justClose);
void flushFeedOut(Feed *fe);

void FeedRSet(FILE *fo);
void FeedCommit(FILE *fo);
void FeedAddDel(FILE *fo, char *gwild, int add);

void FinishRetain(int what);
int QueueRange(const char *label, int *pqnum, int *pqarts, int *pqrun);
const char *ftos(double d);

char	*HName;
char	*NewsPathName;
char	HostName[256];
char	HLabel[256];
int	MaxFds;
int	NumForks;
int	PausedCount;
fd_set	RFds;
Buffer	*PipeAry[MAXFDS];
pid_t	PidAry[MAXFDS];
Feed	*FeBase;
Retain	*ReBase;
volatile int Exiting = 0;
int	DidFork = 0;
int	NewsPort = 119;
struct in_addr NewsBindHost;
int	NumForksExceeded;
time_t	SessionBeg;
int	ArtsEntered;
int	ArtsReceived;
int	ArtsIHave;
int	ArtsCheck;
int	ArtsRej;
int	ArtsStage2Rej;
int	ArtsStage3Rej;
int	ArtsErr;
double	TtlArtsReceived;
double	TtlArtsTested;
double	TtlArtsBytes;
double	TtlArtsFed;
int	LogCount;
int	FeedTableReady;
char	*DebugLabel;
char	*ArgvBuf;
int	ArgvLen;

int
main(int ac, char **av)
{
    char *op = NULL;

    gethostname(HostName, sizeof(HostName));

    rsignal(SIGPIPE, SIG_IGN);
    SessionBeg = time(NULL);

    srandom((long)SessionBeg ^ (getpid() * 100));
    random();
    random();

    NewsBindHost.s_addr = INADDR_ANY;

    /*
     * Options
     */

    {
	int i;

	for (i = 1; i < ac; ++i) {
	    char *ptr = av[i];

	    if (*ptr != '-') {
		if (op) {
		    printf("service option specified twice\n");
		    exit(1);
		}
		op = ptr;
		continue;
	    }
	    ptr += 2;
	    switch(ptr[-1]) {
	    case 's':
		ArgvBuf = ptr - 2;
		ArgvLen = strlen(ArgvBuf);
		memset(ArgvBuf, 0, ArgvLen);
		break;
	    case 'd':
		--ptr;
		while (*ptr == 'd') {
		    ++DebugOpt;
		    ++ptr;
		}
		break;
	    case 'p':
		NewsPathName = (*ptr) ? ptr : av[++i];
		break;
	    case 'P':
		if (*ptr == 0)
		    ptr = av[++i];
		if ((NewsPort = strtol(ptr, NULL, 0)) == 0) {
		    struct servent *sen;
		    if ((sen = getservbyname(ptr, "tcp")) != NULL) {
			NewsPort = ntohl(sen->s_port);
		    } else {
			printf("Unknown service: %s\n", ptr);
			exit(1);
		    }
		}
		break;
	    case 'H':
		if (*ptr == 0)
		    ptr = av[++i];
		if (strtol(ptr, NULL, 0) > 0) {
		    NewsBindHost.s_addr = inet_addr(ptr);
		} else {
		    struct hostent *he;

		    if ((he = gethostbyname(ptr)) != NULL) {
			NewsBindHost = *(struct in_addr *)he->h_addr;
		    } else {
			printf("Unknown host for bindhost option: %s\n", ptr);
			exit(1);
		    }
		}
		break;
	    default:
		printf("unknown option: %s\n", ptr - 2);
		exit(1);
	    }
	}
	if (i > ac) {
	    printf("expected argument to last option\n");
	    exit(1);
	}
    }

    /*
     * For our Path: insertion
     */

    if (NewsPathName == NULL) {
	printf("No '-p newspathname' specified\n");
	exit(1);
    }

    /*
     * The chdir is no longer required, but we do it anyway to
     * have a 'starting point' to look for cores and such.
     */

    if (chdir(SpoolHome) < 0) {
	fprintf(stderr, "chdir('%s'): %s\n", SpoolHome, strerror(errno));
	exit(1);
    }

    if (op == NULL) {
	printf("Must specify service option: (server)\n");
	exit(1);
    } else if (strcmp(op, "server") == 0) {
	DiabloServer();
    } else {
	printf("unknown service option: %s\n", op);
	exit(1);
    }
    return(0);
}

/*
 * This needs fork.  What we are trying to accomplish
 * is to ensure that all the pipes from the children
 * are flushed back to the parent and the parent
 * writes them out before exiting.  Otherwise
 * feed redistribution might fail.
 */

void
sigHup(int sigNo)
{
    Exiting = 1;
    if (DidFork) {
	if (FeedFo) {
	    fflush(FeedFo);
	    fclose(FeedFo);
	}
	exit(1);
    } else {
	if (Exiting == 0) {
	    int i;

	    Exiting = 1;
	    for (i = 0; i < MAXFDS; ++i) {
		if (PidAry[i]) {
		    kill(PidAry[i], SIGHUP);
		}
	    }
	}
    }
}

void
sigUsr1(int sigNo)
{
    ++DebugOpt;
}

void
sigUsr2(int sigNo)
{
    DebugOpt = 0;
}

/*
 * DIABLOSERVER() - The master server process.  It accept()s connections
 *		    and forks off children.
 */

void
DiabloServer(void)
{
    int lfd;
    int ufd;

    /*
     * Detach
     */

    if (DebugOpt == 0) {
	pid_t pid = fork();

	if (pid < 0) {
	    perror("fork");
	    exit(1);
	}
	if (pid > 0) {
	    exit(0);
	}

	/*
	 * Child continues
	 */

	freopen("/dev/null", "w", stdout);
	freopen("/dev/null", "w", stderr);
	freopen("/dev/null", "r", stdin);
	{
	    int fd = open("/dev/tty", O_RDWR);
	    if (fd >= 0) {
		ioctl(fd, TIOCNOTTY, 0);
		close(fd);
	    }
	}
#if USE_SYSV_SETPGRP
	setpgrp();
#else
	setpgrp(0, 0);
#endif
    }

    /*
     * select()/signal() setup
     */

    FD_ZERO(&RFds);

    rsignal(SIGHUP, sigHup);
    rsignal(SIGINT, sigHup);
    rsignal(SIGUSR1, sigUsr1);
    rsignal(SIGUSR2, sigUsr2);

    {
	struct sockaddr_in sin;

	memset(&sin, 0, sizeof(sin));

	openlog("diablo", LOG_PERROR|LOG_PID, LOG_NEWS);

	/*
	 * listen socket for news
	 */

	if ((lfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	    perror("socket");
	    exit(1);
	}
	sin.sin_addr = NewsBindHost;
	sin.sin_port = htons(NewsPort);
	{
	    int on = 1;
	    setsockopt(lfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	    setsockopt(lfd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
	}
	if (bind(lfd, (struct sockaddr *)&sin, sizeof(sin)) < 0) {
	    perror("bind");
	    exit(1);
	}

	fcntl(lfd, F_SETFL, O_NONBLOCK);

	if (listen(lfd, 10) < 0) {
	    perror("listen");
	    exit(1);
	}

	FD_SET(lfd, &RFds);
	if (MaxFds <= lfd)
	    MaxFds = lfd + 1;
    }

    /*
     * UNIX domain socket
     */

    {
	struct sockaddr_un soun;

	memset(&soun, 0, sizeof(soun));

	if ((ufd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0) {
	    perror("udom-socket");
	    exit(1);
	}
	soun.sun_family = AF_UNIX;

	sprintf(soun.sun_path, "%s/.diablosock", NewsHome);
	remove(soun.sun_path);
	if (bind(ufd, (struct sockaddr *)&soun, offsetof(struct sockaddr_un, sun_path[strlen(soun.sun_path)+1])) < 0) {
	    perror("udom-bind");
	    exit(1);
	}
	chmod(soun.sun_path, 0770);

	fcntl(ufd, F_SETFL, O_NONBLOCK);

	if (listen(ufd, 10) < 0) {
	    perror("udom-listen");
	    exit(1);
	}

	FD_SET(ufd, &RFds);
	if (MaxFds <= ufd)
	    MaxFds = ufd + 1;
    }


    /*
     * change my uid/gid
     */

    {
	struct passwd *pw = getpwnam("news");
	struct group *gr = getgrnam("news");
	gid_t gid;

	if (pw == NULL) {
	    perror("getpwnam('news')");
	    exit(1);
	}
	if (gr == NULL) {
	    perror("getgrnam('news')");
	    exit(1);
	}
	gid = gr->gr_gid;
	setgroups(1, &gid);
	setgid(gr->gr_gid);
	setuid(pw->pw_uid);
    }

    while (NumForks || Exiting == 0) {
	fd_set rfds = RFds;
	int i;
	int n;
	struct timeval tv = { 5, 0 };

	if (PausedCount && NumForks == 0) {
	    FinishRetain(RET_PAUSE);
	}

	n = select(MaxFds, &rfds, NULL, NULL, &tv);

	for (i = 0; i < MaxFds; ++i) {
	    if (FD_ISSET(i, &rfds)) {
		--n;
		if (i == lfd)
		    DoAccept(i);
		else if (i == ufd)
		    DoCommand(i);
		else
		    DoPipe(i);
	    }
	}
	{
	    pid_t pid;

	    while ((pid = wait3(NULL, WNOHANG, NULL)) > 0) {
		int i;

		for (i = 0; i < MaxFds; ++i) {
		    if (PidAry[i] == pid) {
			PidAry[i] = 0;
		    }
		}
	    }
	}
	if (NumForks >= MAXFORKS || Exiting) {
	    if (NumForksExceeded == 0) {
		if (Exiting) {
		    syslog(LOG_WARNING, "Exiting, waiting on children");
		} else {
		    syslog(LOG_WARNING, "NumForks exceeded");
		}
		NumForksExceeded = 1;
	    }
	    if (lfd >= 0)
		FD_CLR(lfd, &RFds);
	    if (Exiting) {
		close(lfd);
		lfd = -1;
	    }
	} else {
	    if (NumForksExceeded) {
		if (NumForksExceeded == 1)
		    syslog(LOG_WARNING, "NumForks ok, reaccepting");
		NumForksExceeded = 0;
	    }
	    FD_SET(lfd, &RFds);
	}
    }
    LogSession2();
    FinishRetain(RET_PAUSE);
    FinishRetain(RET_CLOSE);
    flushFeeds(0);
    sleep(2);
}

/*
 * DOACCEPT()	- accept and deal with a new connection
 */

void
DoAccept(int lfd)
{
    int fd;
    struct sockaddr_in asin;
    int alen = sizeof(asin);

    if (Exiting)
	return;

    if ((fd = accept(lfd, (struct sockaddr *)&asin, &alen)) >= 0) {
	int fds[2] = { -1, -1 };
	int ok = 0;

	fcntl(fd, F_SETFL, 0);
	if (pipe(fds) == 0 && fds[0] < MAXFDS) {
	    pid_t pid;

	    /*
	     * we MUST clear our FILE stdio so a call to
	     * exit() doesn't flush FILE structures to
	     * the wrong descriptors!
	     *
	     * WARNING WARNING!  There cannot be a single
	     * critical FILE handle open for which we may
	     * close it's underlying descriptor!
	     */

	    fflush(stdout);
	    fflush(stderr);

	    if ((pid = fork()) == 0) {
		int i;

		flushFeeds(1);	/* close feed descriptors without flushing */

		SessionBeg = time(NULL);

		if (ArgvLen) {
		    sprintf(ArgvBuf, "%-*.*s", ArgvLen, ArgvLen,
			inet_ntoa(asin.sin_addr)
		    );
		}

		closelog();
		openlog("diablo", LOG_PERROR|LOG_PID, LOG_NEWS);

		for (i = 0; i < MaxFds; ++i) {
		    if (i > 2 && i != fds[1] && i != fd) {
			bclose(&PipeAry[fd]);
			close(i);
		    }
		}
		FeedFo = fdopen(fds[1], "w");
		if ((HName = Authenticate(fd, &asin, HLabel)) == NULL) {
		    FILE *fo = fdopen(fd, "w");

		    if (fo == NULL) {
			syslog(LOG_CRIT, "fdopen() of socket failed");
			exit(1);
		    }
		    xfprintf(fo, "502 You have no permission to talk\r\n");
		    syslog(LOG_INFO, "Connection %d from %s (no permission)",
			fds[0],
			inet_ntoa(asin.sin_addr)
		    );
		    exit(0);
		}
		DidFork = 1;

		if (ArgvLen)
		    sprintf(ArgvBuf, "%-*.*s", ArgvLen, ArgvLen, HName);

		syslog(LOG_INFO, "Connection %d from %s %s",
		    fds[0],
		    HName,
		    inet_ntoa(asin.sin_addr)
		);

		/*
		 * Simple session debugging support
		 */
		if (DebugLabel && HLabel[0] && strcmp(DebugLabel, HLabel) == 0) {
		    char path[256];

		    sprintf(path, "/tmp/diablo.debug.%d", (int)getpid());
		    DebugOpt = 1;
		    freopen(path, "a", stderr);
		    freopen(path, "a", stdout);
		    printf("Debug label %s pid %d\n", HLabel, (int)getpid());
		}

		DoSession(fd);
		LogSession();
		syslog(LOG_INFO, "Disconnect %d from %s %s (%d elapsed)",
		    fds[0],
		    HName,
		    inet_ntoa(asin.sin_addr),
		    (int)(time(NULL) - SessionBeg)
		);
		exit(0);
	    }
	    if (pid < 0) {
		syslog(LOG_EMERG, "fork failed: %s", strerror(errno));
	    } else {
		ok = 1;
		++NumForks;
		PidAry[fds[0]] = pid;
	    }
	}
	close(fd);
	if (fds[1] >= 0)
	    close(fds[1]);
	if (fds[0] < 0)
	    syslog(LOG_EMERG, "pipe() failed: %s", strerror(errno));

	if (ok) {
	    fcntl(fds[0], F_SETFL, O_NONBLOCK);
	    FD_SET(fds[0], &RFds);
	    if (MaxFds <= fds[0])
		MaxFds = fds[0] + 1;
	} else {
	    if (fds[0] >= 0) {
		syslog(LOG_WARNING, "Maximum file descriptors exceeded");
		close(fds[0]);
	    }
	}
    }
}

/*
 * DOPIPE()	- handle data returned from our children over a pipe
 */

void
DoPipe(int fd)
{
    char *ptr;
    int maxCount = 100;

    while ((ptr = bgets(&PipeAry[fd], fd)) != NULL && ptr != (void *)-1) {
	char *s1;

	s1 = strtok(ptr, " \t\n");

	if (s1 == NULL)
	    continue;

	if (strcmp(s1, "FEED") == 0) {
	    char *label = strtok(NULL, " \t\n");
	    char *file  = strtok(NULL, " \t\n");
	    char *offsize = strtok(NULL, " \t\n");	/* offset/size */
	    char *msgid = MsgId(strtok(NULL, "\n"));
	    if (label && file && offsize && strcmp(msgid, "<>") != 0)
		writeFeedOut(label, file, msgid, offsize);
	} else if (strncmp(s1, "STATS", 5) == 0) {
	    char *v;
	    int cnt = 0;
	    int bytes = 0;
	    char *mid = "unknown";

	    while ((v = strtok(NULL, " \t\n")) != NULL) {
	       if (v[0] && v[1] == '=') {
		   switch(v[0]) {
		   case 'c':
		       cnt = strtol(v + 2, NULL, 0);
		       break;
		   case 's':
		       bytes = strtol(v + 2, NULL, 0);
		       break;
		   case 'm':
		       mid = v;
		       break;
		    }
	        }
	    }
	    TtlArtsBytes += (double)bytes;
	    TtlArtsFed += (double)cnt;
	    TtlArtsReceived += 1.0;
	    if (++LogCount == 1024) {
		LogCount = 0;
		LogSession2();
	    }
	} else if (strncmp(s1, "FLUSH", 5) == 0) {
	    flushFeeds(0);
	}
	if (--maxCount == 0)
	    break;
    }
    if (ptr == (void *)-1) {
	bclose(&PipeAry[fd]);
	close(fd);
	FD_CLR(fd, &RFds);
	--NumForks;
    }
}

/*
 * WRITEFEEDOUT()	- the master parent writes to the outgoing feed files
 * FLUSHFEEDOUT()
 */

void
writeFeedOut(const char *label, const char *file, const char *msgid, const char *offSize)
{
    Feed *fe;

    /*
     * locate feed
     */

    for (fe = FeBase; fe; fe = fe->fe_Next) {
	if (strcmp(label, fe->fe_Label) == 0)
	    break;
    }

    /*
     * allocate feed if not found
     */

    if (fe == NULL) {
	fe = malloc(sizeof(Feed) + strlen(label) + 1);
	fe->fe_Label = (char *)(fe + 1);
	strcpy(fe->fe_Label, label);
	fe->fe_Fd = xopen(O_APPEND|O_RDWR|O_CREAT, 0644, "%s/dqueue/%s",NewsHome,label);
	if (fe->fe_Fd >= 0) {
	    int bsize;

	    fe->fe_Buf = pagealloc(&bsize);
	    fe->fe_BufMax = bsize;
	    fe->fe_BufIdx = 0;
	    fe->fe_Failed = 0;
	}
	fe->fe_Next = FeBase;
	FeBase = fe;
    }

    /*
     * write to buffered feed, flushing buffers if there is not enough
     * room.  If the line is too long, something has gone wrong and we
     * throw it away.
     *
     * note that we cannot fill the buffer to 100%, because the trailing
     * nul (which we do not write) will overrun it.
     */

    if (fe->fe_Fd >= 0) {
	int l = strlen(file) + strlen(msgid) + strlen(offSize) + 3;
	if (l < fe->fe_BufMax) {
	    if (l >= fe->fe_BufMax - fe->fe_BufIdx)
		flushFeedOut(fe);
	    sprintf(fe->fe_Buf + fe->fe_BufIdx, "%s %s %s\n",
		file, msgid, offSize
	    );
	    fe->fe_BufIdx += l;
	}
    }
}

void
flushFeeds(int justClose)
{
    Feed *fe;

    while ((fe = FeBase) != NULL) {
	if (justClose == 0)
	    flushFeedOut(fe);
	if (fe->fe_Buf)
	    pagefree(fe->fe_Buf);
	if (fe->fe_Fd >= 0)
	    close(fe->fe_Fd);
	fe->fe_Fd = -1;
	fe->fe_Buf = NULL;
	FeBase = fe->fe_Next;
	free(fe);
    }
}

void
flushFeedOut(Feed *fe)
{
    if (fe->fe_BufIdx && fe->fe_Buf && fe->fe_Fd >= 0) {
	/*
	 * flush buffer.  If the write fails, we undo it to ensure
	 * that we do not get garbaged feed files.
	 */
	int n = write(fe->fe_Fd, fe->fe_Buf, fe->fe_BufIdx);
	if (n >= 0 && n != fe->fe_BufIdx) {
	    ftruncate(fe->fe_Fd, lseek(fe->fe_Fd, 0L, 1) - n);
	}
	if (n != fe->fe_BufIdx && fe->fe_Failed == 0) {
	    fe->fe_Failed = 1;
	    syslog(LOG_INFO, "failure writing to feed %s", fe->fe_Label);
	}
    }
    fe->fe_BufIdx = 0;
}

/*
 * AUTHENTICATION()	- authenticate a new connection
 *
 *	If the forward and reverse dns match, look for the
 *	host or ip in diablo.hosts.  If they do not, only
 *	look for the ip in diablo.hosts.
 */

char *
Authenticate(int fd, struct sockaddr_in *asin, char *label)
{
    char *hname = NULL;
    char *pname = NULL;
    int isSecure = 0;
    int isSpoof = 0;
    int isValid = 0;

    /*
     * check reverse lookup verses forward lookup
     */

    {
	struct hostent *he;

	he = gethostbyaddr(
	    (char *)&asin->sin_addr,
	    sizeof(asin->sin_addr),
	    AF_INET
	);

	if (he != NULL) {
	    hname = strcpy(malloc(strlen(he->h_name) + 1), he->h_name);
	    if ((he = gethostbyname(hname)) != NULL) {
		if (strcmp(he->h_name, hname) == 0) {
		    isSecure = 1;
		} else {
		    isSpoof = 1;
		}
	    }

	    /*
	     * Make lowercase for diablo.hosts file
	     */
	    {
		char *p;

		for (p = hname; *p; ++p) {
		    *p = (char)tolower(*p);
		}
	    }
	}
    }
    if (isSecure == 0) {
	if (hname)
	    free(hname);
	hname = NULL;
    }

    /*
     * set pname
     */
    pname = malloc(strlen(inet_ntoa(asin->sin_addr)) + 1);
    sprintf(pname, "%s", inet_ntoa(asin->sin_addr));

    /*
     * Check against diablo.hosts
     */
    {
	FILE *fi = xfopen("r", "%s/diablo.hosts", NewsHome);

	if (fi) {
	    char buf[256];
	    char *l = NULL;

	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		char *s;

		if (buf[0] == '\n' || buf[0] == '#')
		    continue;
		s = strtok(buf, ": \t\r\n");
		if (s && *s) {
		    l = strtok(NULL, ": \t\r\n");	/* label */

		    if (pname && WildCmp(s, pname) == 0) {
			isValid = 1;
			break;
		    }
		    if (hname && WildCmp(s, hname) == 0) {
			isValid = 1;
			break;
		    }
		}
	    }
	    if (isValid) {
		if (l && strlen(l) < 256)
		    strcpy(label, l);
		else
		    label[0] = 0;
	    }
	    fclose(fi);
	} else {
	    syslog(LOG_INFO, "%s/diablo.hosts file not found", NewsHome);
	}
    }

    if (isValid) {
	if (hname == NULL) {
	    hname = pname;
	    pname = NULL;
	}
    } else {
	if (hname)
	    free(hname);
	hname = NULL;
	if (isSpoof) {
	    syslog(
		LOG_EMERG, 
		"DNS failure or spoof from %s", pname
	    );
	}
    }
    if (pname) {
	free(pname);
	pname = NULL;
    }
    return(hname);
}

/*
 * DOSESSION()	- a child process to handle a diablo connection
 */

void
DoSession(int fd)
{
    FILE *fi;
    FILE *fo;
    char buf[MAXLINE];
    int streamMode = 0;
    int syntax = 0;
    int unimp = 0;

    {
	int nfd = dup(fd);
	if (nfd < 0) {
	    syslog(LOG_CRIT, "DoSession: dup()");
	    exit(1);
	}
	fi = fdopen(nfd, "r");
	fo = fdopen(fd, "w");
	if (fi == NULL || fo == NULL) {
	    syslog(LOG_CRIT, "DoSession() fdopen failure");
	    exit(1);
	}
    }
    if (PausedCount) {
	xfprintf(fo, "502 %s DIABLO is currently paused\r\n", HostName);
	fflush(fo);
	exit(0);
    }


    HistoryOpen(NULL, 0);
    LoadNewsFeed(NULL, 1);
    LoadExpireCtl(NULL, 1);

    xfprintf(fo, "200 %s News server DIABLO %4.2f ready\r\n", HostName, VERS);
    fflush(fo);

    while (fgets(buf, sizeof(buf), fi) != NULL) {
	char *cmd;

	if (DebugOpt)
	    dprintf("%d << %s", (int)getpid(), buf);

	cmd = strtok(buf, " \t\r\n");

	if (cmd == NULL)
	    continue;

	if (strcasecmp(cmd, "ihave") == 0) {
	    char *msgid = MsgId(strtok(NULL, "\r\n"));

	    ++ArtsIHave;
	    TtlArtsTested += 1.0;
	    if (strcmp(msgid, "<>") == 0 || HistoryLookup(msgid, NULL, NULL, NULL) == 0) {
		xfprintf(fo, "435 %s\r\n", msgid);
	    } else {
		int r;

		xfprintf(fo, "335 %s\r\n", msgid);
		fflush(fo);
		switch((r = LoadArticle(fi, buf, msgid, 0))) {
		case RCOK:
		    xfprintf(fo, "235\r\n");	/* article posted ok	*/
		    break;
		case RCALREADY:
		    xfprintf(fo, "435\r\n");	/* already have it	*/
		    break;
		case RCTRYAGAIN:
		    xfprintf(fo, "436 I/O error, try again later\r\n");
		    break;
		case RCREJECT:
		    xfprintf(fo, "437 Rejected\r\n");
		    break;
		case RCERROR:
		    ++ArtsErr;
		    /*
		     * protocol error during transfer (e.g. no terminating .)
		     */
		    xfprintf(fo, "436 Protocol error, missing terminator\r\n");
		    break;
		default:
		    ++ArtsErr;
		    /*
		     * An I/O error of some sort (e.g. disk full).
		     */
		    xfprintf(fo, "400 File Error: %s\r\n", strerror(-r));
                    syslog(LOG_ERR, "%-20s %s code 400 File Error: %s",
                        HName,
                        msgid,
                        strerror(-r)
                    ); 
		    LogSession();
		    exit(1);
		    break;	/* not reached */
		}
	    }
	} else if (strcasecmp(cmd, "check") == 0) {
	    char *msgid = MsgId(strtok(NULL, "\r\n"));

	    ++ArtsCheck;
	    TtlArtsTested += 1.0;
	    if (strcmp(msgid, "<>") == 0 || HistoryLookup(msgid, NULL, NULL, NULL) == 0) {
		xfprintf(fo, "438 %s\r\n", msgid);
	    } else {
		xfprintf(fo, "238 %s\r\n", msgid);
	    }
	} else if (strcasecmp(cmd, "takethis") == 0) {
	    char *msgid = MsgId(strtok(NULL, "\r\n"));
	    int r;
	    int alreadyResponded = 0;

	    TtlArtsTested += 1.0;
	    if (HistoryLookup(msgid, NULL, NULL, NULL) == 0) {
		xfprintf(fo, "439 %s\r\n", msgid);
		fflush(fo);
		LoadArticle(fi, buf, msgid, 1);
		r = RCALREADY;
		++ArtsStage2Rej;
		alreadyResponded = 1;
	    } else {
		r = LoadArticle(fi, buf, msgid, 0);
	    }

	    if (alreadyResponded == 0) {
		switch(r) {
		case RCOK:
		    xfprintf(fo, "239 %s\r\n", msgid);	/* thank you */
		    break;
		case RCALREADY:
		    xfprintf(fo, "439 %s\r\n", msgid);	/* already have it or do not requeue it	*/
		    break;
		case RCTRYAGAIN:
		    xfprintf(fo, "431 %s\r\n", msgid);
		    break;
		case RCREJECT:
		    xfprintf(fo, "439 %s\r\n", msgid);
		    break;
		case RCERROR:
		    xfprintf(fo, "431 %s\r\n", msgid);	/* article failed due to something, do not req */
		    ++ArtsErr;
		    break;
		default:
		    ++ArtsErr;
		    /*
		     * An I/O error of some sort (e.g. disk full).
		     */
		    xfprintf(fo, "400 File Error: %s, %s\r\n", msgid, strerror(-r));
		    syslog(LOG_ERR, "%-20s %s code 400 File Error: %s",
			HName,
			msgid,
			strerror(-r)
		    ); 
		    LogSession();
		    exit(1);
		    break;	/* not reached */
		}
	    }
	} else if (strcasecmp(cmd, "head") == 0) {
	    char *msgid = MsgId(strtok(NULL, "\r\n"));
	    char *data = NULL;
	    long fsize = 0;
	    int pmart = 0;

	    if (strcmp(msgid, "<>") == 0) {
		xfprintf(fo, "443 Bad Message-ID\r\n");
	    } else if (HistoryLookup(msgid, &data, &fsize, &pmart) == 0) {
		if (data) {
		    int b;
		    int i;

		    xfprintf(fo, "221 0 head %s\r\n", msgid);

		    if (DebugOpt)
			dprintf(">> (DATA)\n");

		    for (i = b = 0; i < fsize; b = i) {
			/*
			 * if first character is a ., escape it
			 */

			if (data[i] == '.')
			    fputc('.', fo);
			while (i < fsize && data[i] != '\n')
			    ++i;
			fwrite(data + b, i - b, 1, fo);

			/*
			 * if i - b == 0 we are at the end of the headers.  We
			 * do not print the blank line.
			 */
			if (i - b == 0)
			    break;

			fputc('\r', fo);
			fputc('\n', fo);
			++i;

			/*
			 * if i > fsize, we hit the end of the file without
			 * a terminating LF.  We don't have to do anything
			 * since we've already terminated the last line.
			 */
		    }
		    xfprintf(fo, ".\r\n");
		} else {
		    xfprintf(fo, "430\r\n");
		    if (DebugOpt)
			dprintf(">> (NO DATA: UNABLE TO FIND ARTICLE)\n");
		}
	    } else {
		xfprintf(fo, "430\r\n");
	    }
	    if (data)
		xunmap(data, fsize + pmart);
	} else if (strcasecmp(cmd, "stat") == 0) {
	    char *msgid = MsgId(strtok(NULL, "\r\n"));

	    if (strcmp(msgid, "<>") == 0) {
		xfprintf(fo, "443 Bad Message-ID\r\n");
	    } else if (HistoryLookup(msgid, NULL, NULL, NULL) == 0) {
		xfprintf(fo, "223 0 %s\r\n", msgid);
	    } else {
		xfprintf(fo, "430\r\n");
	    }
	} else if (strcasecmp(cmd, "feedrset") == 0) {
	    if (HLabel[0]) {
		FeedRSet(fo);
	    } else {
		xfprintf(fo, "490 Operation not allowed\r\n");
	    }
	} else if (strcasecmp(cmd, "feedcommit") == 0) {
	    if (HLabel[0]) {
		if (FeedTableReady)
		    FeedCommit(fo);
		else
		    xfprintf(fo, "491 No feedrset/add/del\r\n");
	} else {
	    xfprintf(fo, "490 Operation not allowed\r\n");
	    }
	} else if (strcasecmp(cmd, "feedadd") == 0) {
	    char *p = strtok(NULL, " \t\r\n");
	    if (p) {
		if (HLabel[0]) {
		    if (FeedTableReady)
			FeedAddDel(fo, p, 1);
		    else
			xfprintf(fo, "491 No feedrset\r\n");
		} else {
		    xfprintf(fo, "490 Operation not allowed\r\n");
		}
	    } else {
		xfprintf(fo, "491 Syntax Error\r\n");
	    }
	} else if (strcasecmp(cmd, "feeddel") == 0) {
	    char *p = strtok(NULL, " \t\r\n");
	    if (p) {
		if (HLabel[0]) {
		    if (FeedTableReady)
			FeedAddDel(fo, p, 0);
		    else
			xfprintf(fo, "491 No feedrset\r\n");
		} else {
		    xfprintf(fo, "490 Operation not allowed\r\n");
		}
	    } else {
		xfprintf(fo, "491 Syntax Error\r\n");
	    }
	} else if (strcasecmp(cmd, "mode") == 0) {
	    char *p = strtok(NULL, " \t\r\n");
	    if (p && strcasecmp(p, "stream") == 0) {
		streamMode = 1;
		xfprintf(fo, "203 StreamOK.\r\n");
	    } else if (p && strcasecmp(p, "reader") == 0) {
		unimp = 1;
	    } else {
		syntax = 1;
	    }
	} else if (strcasecmp(cmd, "outq") == 0) {
	    int qnum;
	    int qarts;
	    int qrun;

	    if (HLabel[0] && QueueRange(HLabel, &qnum, &qarts, &qrun) == 0) {
		xfprintf(fo, "290 qfile-backlog=%d arts=%d now-running=%d\r\n", 
		    qnum,
		    qarts,
		    qrun
		);
	    } else {
		xfprintf(fo, "491 No queue info available\r\n");
	    }
	} else if (strcasecmp(cmd, "authinfo") == 0) {
	    xfprintf(fo, "281 Authentication ok, no authentication required\r\n");
	} else if (strcasecmp(cmd, "quit") == 0) {
	    xfprintf(fo, "205 See ya later.\r\n");
	    fflush(fo);
	    break;
	} else if (strcasecmp(cmd, "help") == 0) {
	    xfprintf(fo, "100 Legal commands\r\n");
	    xfprintf(fo, "\tauthinfo\r\n"
			"\thelp\r\n"
			"\tihave\r\n"
			"\tcheck\r\n"
			"\ttakethis\r\n"
			"\tmode\r\n"
			"\tquit\r\n"
			"\thead\r\n"
			"\tstat\r\n"
			"\toutq\r\n"
			"\tfeedrset\r\n"
			"\tfeedadd grpwildcard\r\n"
			"\tfeeddel grpwildcard\r\n"
			"\tfeedcommit\r\n"
	    );
	    xfprintf(fo, ".\r\n");
	} else {
	    syntax = 1;
	    if (strcasecmp(cmd, "list") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "article") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "body") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "group") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "last") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "newgroups") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "newnews") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "next") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "post") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "slave") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "xhdr") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "xpath") == 0)
		unimp = 1;
	    if (strcasecmp(cmd, "xreplic") == 0)
		unimp = 1;
	    if (unimp)
		syntax = 0;
	}
	if (unimp) {
	    xfprintf(fo, "500 command not implemented\r\n");
	    unimp = 0;
	}
	if (syntax) {
	    xfprintf(fo, "500 Syntax error or bad command\r\n");
	    syntax = 0;
	}
	fflush(fo);
	if (ArgvLen > 32) {
	    char tbuf[256];

	    sprintf(tbuf, "ihav=%-4d chk=%-4d rec=%-4d ent=%-4d %s",
		ArtsIHave,
		ArtsCheck,
		ArtsReceived,
		ArtsEntered,
		HName
	    );
	    sprintf(ArgvBuf, "%-*.*s", ArgvLen, ArgvLen, tbuf);
	}
    }
    fclose(fi);
    fclose(fo);
}

/*
 * LOADARTICLE()	- get an article from the remote
 */

static int ArtFd = -1;
static unsigned short ArtIter;
char       ArtPath[256];

int
LoadArticle(FILE *fi, char *buf, const char *msgid, int noWrite)
{
    /*
     * create article file
     */
    int r = RCOK;
    int size = 0;
    int haveValidFile = 0;
    long bpos = 0;
    char path[256];
    History h = { 0 };

    h.hv = hhash(msgid);
    h.gmt = time(NULL) / 60;

    path[0] = 0;

    /*
     * Obtain the file used to handle this article.  The file may already
     * be cached... generally, we wind up with one new file every 10 minutes.
     * to the file at the same time, we will be ok.
     */

    errno = 0;	/* in case noWrite is true */

    h.boffset = 1;	/* force new queue file format in ArticleFileName */
    h.bsize = 1;

    h.iter = ArtIter;
    ArticleFileName(path, &h, 1);

    if (noWrite == 0 && (ArtFd < 0 || strcmp(path, ArtPath) != 0)) {
	int cnt;

	FlushArtFd();

	for (cnt = 0; cnt < 100000; ++cnt) {
	    h.iter = random() & 32767;

	    errno = 0;

	    ArticleFileName(path, &h, 1);
	    if ((ArtFd = open(cdcache(path), O_RDWR|O_CREAT, 0644)) >= 0) {
		struct stat st;

		if (xflock(ArtFd, XLOCK_EX|XLOCK_NB) < 0 || (fstat(ArtFd, &st), st.st_nlink) == 0) {
		    close(ArtFd);
		    errno = 0;
		    ArtFd = -1;
		    continue;
		}
		strcpy(ArtPath, path);
		ArtIter = h.iter;
		break;
	    }

	    /*
	     * If there's something wrong with us, break out, otherwise
	     * attemp to make the directory and continue.  The sleep
	     * is a fail-safe.
	     */

	    sleep(1);
	    if (errno == ENOSPC || errno == EIO || errno == EMFILE || 
		errno == ENFILE
	    ) {
		break;
	    }

	    /*
	     * attempt to make the missing directory, breaking out if we have
	     * run out of space, looping otherwise.  The sleep is a failsafe.
	     */

	    *strrchr(path, '/') = 0;
	    if (mkdir(path, 0755) < 0) {
		if (errno == ENOSPC || errno == EIO)
		    break;
	    }
	    /* go on to the next iteration	*/
	}
    }

    /*
     * If errno is 0 here, we usually want
     * to RCTRYAGAIN.  However, the client
     * may not do the right thing so we
     * return RCALREADY instead.
     */

    if (noWrite || ArtFd < 0) {
	r = -errno;
	if (r >= 0)
	    r = RCALREADY;
    } else {
	haveValidFile = 1;
	bpos = lseek(ArtFd, 0L, 2);
    }

    /*
     * r will be RCOK if we have a valid file, RCTRYAGAIN if we do not
     *
     * Now scan the file.  Set r to RCERROR if it is RCOK.  It will be
     * set back to RCOK when we get the ., otherwise it's an error.
     */

    if (r == RCOK)
	r = RCERROR;

    {
	FILE *fo = NULL;
	int inHeader = 1;
	int skipHeader = 0;
	int haveMsgId = 0;
	int haveDate = 0;
	char nglist[MAXLINE];
	char npath[MAXLINE];
	char dateBuf[64];

	nglist[0] = 0;
	npath[0] = 0;
	dateBuf[0] = 0;

	if (noWrite == 0 && ArtFd >= 0)
	    fo = fdopen(ArtFd, "w");

	while (fgets(buf, MAXLINE, fi) != NULL) {
	    int l = strlen(buf);
	    char *p = buf;

	    size += l + 1;

	    /*
	     * Remove CRLF (XXX)
	     */

	    if (l >= 2) {
		if (p[l-2] == '\r')
		    p[l-2] = 0;
		if (p[l-1] == '\n')
		    p[l-1] = 0;
	    }

	    /*
	     * Look for end of article
	     */
	    if (strcmp(p, ".") == 0) {
		if (r == RCERROR)
		    r = RCOK;
		break;
	    }

	    /*
	     * Look for dot escape
	     */
	    if (*p == '.')
		++p;

	    /*
	     * Extract certain header information
	     */

	    if (inHeader) {
		/*
		 * if skipping a header-in-progress
		 */
		if (skipHeader) {
		    if (p[0] == ' ' || p[0] == '\t')
			continue;
		    skipHeader = 0;
		}

		/*
		 * prepend to Path, remove Xref lines
		 */

		if (strncasecmp(p, "Path:", 5) == 0) {
		    p = p + 5;
		    while (*p == ' ' || *p == '\t')
			++p;
		    if (fo)
			fprintf(fo, "Path: %s!", NewsPathName);
		    strcpy(npath, p);
		} else if (strncasecmp(p, "Xref:", 5) == 0) {
		    skipHeader = 1;
		    continue;
		} else if (strncasecmp(p, "Newsgroups:", 11) == 0) {
		    strcpy(nglist, p + 11);
		} else if (strncasecmp(p, "Date:", 5) == 0) {
		    haveDate = 1;
		    strncpy(dateBuf, p + 5, sizeof(dateBuf) - 1);
		    dateBuf[sizeof(dateBuf)-1] = 0;
		} else if (strncasecmp(p, "Message-ID:", 11) == 0) {
		    char *ps = strchr(p + 11, '<');
		    char *pe = ps ? strchr(ps, '>') : NULL;

		    if (ps && pe) {
			int l = pe - ps + 1;
			if (strlen(msgid) == l && strncmp(msgid, ps, l) == 0)
			    haveMsgId = 1;
		    }
		    if (haveMsgId == 0) {
			syslog(LOG_ERR, "%-20s message-id mismatch, command: %s, article: %s",
			    HName,
			    msgid,
			    ((ps) ? ps : "(null)")
			);
		    }
		} else if (*p == 0) {
		    /*
		     * end of header
		     */
		    inHeader = 0;
		}
	    }
	    if (fo)
		fprintf(fo, "%s\n", p);
	}
	if (fo) {
	    h.boffset = bpos;
	    h.bsize = ftell(fo) - bpos;
	    fputc(0, fo);		/* terminator (sanity check) */
	    fflush(fo);
	}

	/*
	 * check output file for error, only if r == RCOK.
	 * If there was a problem, change the return code
	 * to RCTRYAGAIN.
	 */

	if (fo != NULL && ferror(fo)) {
	    if (r == RCOK) {
		sleep(1);		/* failsafe */
		r = -errno;
		if (r >= 0)
		    r = RCTRYAGAIN;
	    }
	}

	/*
	 * If the message isn't already error'd out,
	 * reject it.
	 */

	if (nglist[0] == 0 || haveMsgId == 0) {
	    if (r >= 0 && r != RCERROR)
		r = RCREJECT;
	}

	/*
	 * If the messsage is too old, reject it.  The standard
	 * is 16 days, but our parsedate() routine is only 
	 * an estimate, so we use 14. 
	 *
	 * XXX reject the article if parsedate doesn't have a clue ?
	 */
	if (r == RCOK && haveDate) {
	    time_t t = parsedate(dateBuf);

	    if (t != (time_t)-1) {
		long dt = time(NULL) - t;

		if (dt > 14 * 24 * 60 * 60 || dt < - 14 * 24 * 60 * 60)
		    r = RCREJECT;
	    }
	}

	if (fo) {
	    /*
	     * force an illegal descriptor because we will loose our
	     * flock if we close a dup(), so we can't use a dup.
	     */
	    fileno(fo) = -1;
	    fclose(fo);
	}

	if (r == RCOK && IsFiltered(HLabel, nglist) == 0)
	    r = RCREJECT;

	/*
	 * If everything is ok, commit the message
	 */

	if (r == RCOK) {
#if USE_SYSV_SIGNALS
	    sighold(SIGHUP);
#else
	    int smask = sigblock((1 << SIGHUP));
#endif
	    /*
	     * Get the expiration.  An expiration of zero
	     * is an indication that we should not take 
	     * the article (usually due to size), but
	     * we need to add it to the history file 
	     * so we can reject it later.
	     */

	    h.exp = GetExpire(msgid, &h, nglist, size);

	    r = HistoryAdd(msgid, &h);

	    if (r != RCOK)
		++ArtsStage3Rej;

	    if (h.exp == 0)
		r = RCREJECT;

	    if (r == RCOK) {
		if (FeedAdd(msgid, &h, nglist, npath, size) < 0) {
		    /*
		     * If we loose our pipe, exit immediately.
		     */
		    exit(1);
		}
	    }
#if USE_SYSV_SIGNALS
	    sigrelse(SIGHUP);
#else
	    sigsetmask(smask);
#endif
	}
	++ArtsReceived;
	if (r == RCOK)
	    ++ArtsEntered;
	else
	    ++ArtsRej;

	if ((ArtsReceived % 1024) == 0)
	    LogSession();
    }

    /*
     * If we created a file but the return code is
     * not RCOK, remove the file.
     */

    if (r != RCOK && haveValidFile) {
	if (ArtFd >= 0)
	    ftruncate(ArtFd, bpos);
    }

    return(r);
}

void
FlushArtFd(void)
{
    if (ArtFd >= 0) {
	if (lseek(ArtFd, 0L, 2) == 0)	/* empty */
	    remove(ArtPath);
	close(ArtFd);
	ArtFd = -1;
    }
}


/*
 * LOGSESSION()	- Log statistics for a session
 */

void
LogSession(void)
{
    time_t t = time(NULL);
    long dt = t - SessionBeg;
    long nuse;

    nuse = ArtsCheck + ArtsIHave;
    if (nuse < ArtsReceived)
	nuse = ArtsReceived;

    syslog(LOG_INFO, "%-20s secs=%-4d ihave=%-4d chk=%-4d rec=%-4d rej=%-4d predup=%-4d posdup=%-4d err=%-4d added=%-4d (%d/sec)",
	HName,
	dt,
	ArtsIHave,
	ArtsCheck,
	ArtsReceived,
	ArtsRej,
	ArtsStage2Rej,
	ArtsStage3Rej,
	ArtsErr,
	ArtsEntered,
	nuse / ((dt == 0) ? 1 : dt)
    );
    ArtsIHave = 0;
    ArtsCheck = 0;
    ArtsReceived = 0;
    ArtsEntered = 0;
    ArtsRej = 0;
    ArtsStage2Rej = 0;
    ArtsStage3Rej = 0;
    ArtsErr = 0;
    SessionBeg = t;
}

void
LogSession2(void)
{
    int dt = (int)(time(NULL) - SessionBeg);

    syslog(LOG_INFO, 
	"DIABLO uptime=%d:%02d arts=%s tested=%s bytes=%s fed=%s",
	dt / 3600,
	dt / 60 % 60,
	ftos(TtlArtsReceived),
	ftos(TtlArtsTested),
	ftos(TtlArtsBytes),
	ftos(TtlArtsFed)
    );
}

/*
 * DOCOMMAND() - handles a control connection
 */

void
DoCommand(int ufd)
{
    int fd;
    int retain = 0;
    struct sockaddr_in asin;
    int alen = sizeof(asin);

    while ((fd = accept(ufd, (struct sockaddr *)&asin, &alen)) > 0) {
	FILE *fi;
	FILE *fo;

	fcntl(fd, F_SETFL, 0);
	fo = fdopen(dup(fd), "w");
	fi = fdopen(fd, "r");

	if (fi && fo) {
	    char buf[MAXLINE];
	    while (fgets(buf, sizeof(buf), fi) != NULL) {
		char *s1 = strtok(buf, " \t\r\n");

		if (s1 == NULL)
		    continue;

		if (strcmp(s1, "flush") == 0) {
		    fprintf(fo, "211 Flushing feeds\n");
		    fflush(fo);
		    flushFeeds(0);
		} else if (strcmp(s1, "pause") == 0) {
		    retain = RET_PAUSE;
		    ++PausedCount;
		    fprintf(fo, "200 Pause, count %d\n", PausedCount);
		    if (PausedCount == 1) {
			int i;

			for (i = 0; i < MAXFDS; ++i) {
			    if (PidAry[i]) {
				kill(PidAry[i], SIGHUP);
			    }
			}
		    }
		} else if (strcmp(s1, "go") == 0) {
		    if (PausedCount)
			--PausedCount;
		    fprintf(fo, "200 Resume, count %d\n", PausedCount);
		} else if (strcmp(s1, "quit") == 0) {
		    fprintf(fo, "200 Goodbye\n");
		    fprintf(fo, ".\n");
		    break;
		} else if (strcmp(s1, "exit") == 0 || strcmp(s1, "aexit")==0) {
		    int i;

		    if (s1[0] == 'e')
			retain = RET_CLOSE;

		    fprintf(fo, "211 Exiting\n");
		    fflush(fo);

		    if (Exiting == 0) {
			Exiting = 1;
			for (i = 0; i < MAXFDS; ++i) {
			    if (PidAry[i]) {
				kill(PidAry[i], SIGHUP);
			    }
			}
		    } else {
			fprintf(fo, "200 Exit is already in progress\n");
		    }
		    flushFeeds(0);
		} else if (strcmp(s1, "debug") == 0) {
		    char *s2 = strtok(NULL, " \t\r\n");

		    if (DebugLabel)
			free(DebugLabel);

		    if (s2) {
			DebugLabel = strcpy(malloc(strlen(s2) + 1), s2);
			fprintf(fo, "200 Debugging '%s'\n", DebugLabel);
		    } else {
			DebugLabel = NULL;
			fprintf(fo, "200 Debugging turned off\n");
		    }
		} else {
		    fprintf(fo, "500 Unrecognized command: %s\n", s1);
		}
		if (retain == 0)
		    fprintf(fo, ".\n");
		fflush(fo);
		if (retain)
		    break;
	    }
	}
	if (retain) {
	    Retain *ret = malloc(sizeof(Retain));

	    ret->re_Next = ReBase;
	    ret->re_Fi = fi;
	    ret->re_Fo = fo;
	    ret->re_What = retain;
	    ReBase = ret;
	} else {
	    if (fi)
		fclose(fi);
	    if (fo)
		fclose(fo);
	}
    }
}

/*
 * REMOTE FEED CODE
 */

FILE *Ft;
int  FtNumCommit;

void
FeedRSet(FILE *fo)
{
    Ft = xfopen("w", "%s/feeds/%s.new", NewsHome, HLabel);
    if (Ft != NULL) {
	FeedTableReady = 1;
	FtNumCommit = 0;
	xfprintf(fo, "290 Label Reset\r\n");
    } else {
	xfprintf(fo, "490 file create failed\r\n");
    }
}

void
FeedCommit(FILE *fo)
{
    char path1[256+64];
    char path2[256+64];

    sprintf(path1, "%s/feeds/%s.new", NewsHome, HLabel);
    sprintf(path2, "%s/feeds/%s", NewsHome, HLabel);
    fflush(Ft);
    if (ferror(Ft)) {
	remove(path1);
	xfprintf(fo, "490 file write failed\r\n");
    } else {
	if (FtNumCommit == 0) {
	    xfprintf(fo, "290 empty feed list, reverting to initial\r\n");
	    remove(path1);
	    remove(path2);
	} else {
	    xfprintf(fo, "290 feed commit complete\r\n");
	    rename(path1, path2);
	    TouchNewsFeed();
	}
    }
    fclose(Ft);
    FeedTableReady = 0;
}

void
FeedAddDel(FILE *fo, char *gwild, int add)
{
    int i;

    if (strlen(gwild) > MAXGNAME - 8) {
	xfprintf(fo, "490 wildcard too long\r\n");
	return;
    }
    if (FtNumCommit == MAXFEEDTABLE) {
	xfprintf(fo, "490 too many entries, max is %d\r\n", MAXFEEDTABLE);
	return;
    }
    for (i = 0; gwild[i]; ++i) {
	if (isalnum(gwild[i]))
	    continue;
	if (gwild[i] == '*')
	    continue;
	if (gwild[i] == '?')
	    continue;
	if (gwild[i] == '.')
	    continue;
	if (gwild[i] == '-')
	    continue;
	xfprintf(fo, "490 illegal character: %c\r\n", gwild[i]);
	break;
    }
    if (gwild[i] == 0) {
	xfprintf(fo, "290 ok\r\n");
	fprintf(Ft, "%s\t%s\n", ((add) ? "addgroup" : "delgroup"), gwild);
	++FtNumCommit;
    }
}

void
FinishRetain(int what)
{
    Retain *ret;
    Retain **pret;

    for (pret = &ReBase; (ret = *pret) != NULL; ) {
	if (ret->re_What == what) {
	    if (ret->re_Fo)
		fprintf(ret->re_Fo, "200 Operation Complete\n.\n");
	    *pret = ret->re_Next;
	    if (ret->re_Fi)
		fclose(ret->re_Fi);
	    if (ret->re_Fo)
		fclose(ret->re_Fo);
	    free(ret);
	} else {
	    pret = &ret->re_Next;
	}
    }
}

int
QueueRange(const char *label, int *pqnum, int *pqarts, int *pqrun)
{
    FILE *fi = xfopen("r", "%s/dqueue/.%s.seq", NewsHome, label);
    int r = -1;

    if (fi) {
	int qbeg;
	int qend;
	if (fscanf(fi, "%d %d", &qbeg, &qend) == 2) {
	    r = 0;
	    *pqrun = 0;
	    *pqarts = -1;
	    *pqnum = qend - qbeg;

	    if (*pqnum < 500000) {
		int i;

		*pqarts = 0;
		for (i = qbeg; i < qend; ++i) {
		    FILE *fj = xfopen("r", "%s/dqueue/%s.S%05d", NewsHome, label, i);
		    if (fj) {
			char buf[256];

			while (fgets(buf, sizeof(buf), fj) != NULL) {
			    ++*pqarts;
			}
			if (xflock(fileno(fj), XLOCK_EX|XLOCK_NB) < 0)
			    ++*pqrun;
			fclose(fj);
		    }
		}
	    }
	}
	fclose(fi);
    }
    return(r);
}

#define ONE_K	1024.0	
#define ONE_M	(1024.0*1024.0)
#define ONE_G	(1024.0*1024.0*1024.0)

const char *
ftos(double d)
{
    static char FBuf[8][32];
    static int FCnt;
    char *p = FBuf[FCnt];

    if (d < 1024.0) {
	sprintf(p, "%d", (int)d);
    } else if (d < ONE_M) {
	sprintf(p, "%d.%03dK", (int)(d / ONE_K), (int)d % (int)ONE_K);
    } else if (d < ONE_G) {
	sprintf(p, "%d.%03dM", (int)(d / ONE_M), (int)(d / ONE_K) % (int)ONE_K);
    } else {
	sprintf(p, "%d.%03dG", (int)(d / ONE_G), (int)(d / ONE_M) % (int)ONE_K);
    }

    FCnt = (FCnt + 1) & 7;
    return(p);
}

